import numpy as np
import pandas as pd
import pytest

from pandabear.column_checks import (
    series_greater,
    series_greater_equal,
    series_isin,
    series_less,
    series_less_equal,
    series_notin,
    series_notnull,
    series_str_contains,
    series_str_endswith,
    series_str_startswith,
)
from pandabear.exceptions import ColumnCheckError


def test_series_greater_equal():
    assert not series_greater_equal(pd.Series([1, 2, 3]), 2).all()
    assert series_greater_equal(pd.Series([1, 2, 3]), 1).all()
    assert series_greater_equal(pd.Series([1, 2, 3]), 0).all()
    # and for some strings
    assert not series_greater_equal(pd.Series(["a", "b", "c"]), "b").all()
    assert series_greater_equal(pd.Series(["a", "b", "c"]), "a").all()


def test_series_greater():
    assert not series_greater(pd.Series([1, 2, 3]), 2).all()
    assert not series_greater(pd.Series([1, 2, 3]), 1).all()
    assert series_greater(pd.Series([1, 2, 3]), 0).all()
    # and for some strings
    assert not series_greater(pd.Series(["a", "b", "c"]), "b").all()
    assert not series_greater(pd.Series(["a", "b", "c"]), "a").all()
    assert series_greater(pd.Series(["a", "b", "c"]), "0").all()


def test_series_less_equal():
    assert not series_less_equal(pd.Series([1, 2, 3]), 2).all()
    assert series_less_equal(pd.Series([1, 2, 3]), 3).all()
    assert series_less_equal(pd.Series([1, 2, 3]), 4).all()
    # and for some strings
    assert series_less_equal(pd.Series(["a", "b", "c"]), "c").all()
    assert not series_less_equal(pd.Series(["a", "b", "c"]), "a").all()


def test_series_less():
    assert series_less(pd.Series([1, 2, 3]), 4).all()
    assert not series_less(pd.Series([1, 2, 3]), 1).all()
    assert not series_less(pd.Series([1, 2, 3]), 0).all()
    # and for some strings
    assert series_less(pd.Series(["a", "b", "c"]), "d").all()
    assert not series_less(pd.Series(["a", "b", "c"]), "a").all()
    assert not series_less(pd.Series(["a", "b", "c"]), "0").all()


def test_series_isin():
    assert not series_isin(pd.Series([1, 2, 3]), [2]).all()
    assert series_isin(pd.Series([1, 2, 2]), [1, 2]).all()
    # and for some strings
    assert not series_isin(pd.Series(["a", "b", "c"]), ["b"]).all()
    assert series_isin(pd.Series(["a", "b", "b"]), ["a", "b"]).all()


def test_series_notin():
    assert series_notin(pd.Series([1, 2, 3]), [4]).all()
    assert not series_notin(pd.Series([1, 2, 2]), [2]).all()
    # and for some strings
    assert series_notin(pd.Series(["a", "b", "c"]), ["d", "e"]).all()
    assert not series_notin(pd.Series(["a", "b", "b"]), ["a", "b"]).all()


def test_series_str_contains():
    assert not series_str_contains(pd.Series(["a", "b", "c"]), "b").all()
    assert series_str_contains(pd.Series(["ab", "bb", "cb"]), "b").all()
    assert not series_str_contains(pd.Series(["a", "b", "c"]), "0").all()


def test_series_str_endswith():
    assert not series_str_endswith(pd.Series(["a", "b", "c"]), "b").all()
    assert series_str_endswith(pd.Series(["ac", "bc", "cc"]), "c").all()
    assert not series_str_endswith(pd.Series(["a", "b", "c"]), "0").all()


def test_series_str_startswith():
    assert not series_str_startswith(pd.Series(["a", "b", "c"]), "b").all()
    assert series_str_startswith(pd.Series(["ca", "cb", "cc"]), "c").all()


def test_series_notnull():
    assert series_notnull(pd.Series([1, 2, 3])).all()
    assert not series_notnull(pd.Series([1, 2, 3, None])).all()
    assert not series_notnull(pd.Series([1.0, 2.0, np.nan])).all()
    assert not series_notnull(pd.Series([pd.NaT, pd.Timestamp("1939-05-27")])).all()


def test_ColumnCheckError():
    check_func = series_greater
    series = pd.Series([1, 2, 3], name="test")

    # 1. test failure and message with info on failing rows
    expected_message = r"Column 'test' failed check greater\(2\): 2 of 3 \(67 %\)"
    with pytest.raises(ColumnCheckError, match=expected_message):
        result = check_func(series, 2)
        raise ColumnCheckError(check_name=check_func.__name__, check_value=2, series=series, result=result)


if __name__ == "__main__":
    test_series_greater_equal()
    test_series_greater()
    test_series_less_equal()
    test_series_less()
    test_series_isin()
    test_series_notin()
    test_series_str_contains()
    test_series_str_endswith()
    test_series_str_startswith()
    test_series_notnull()
    test_check_handler()



from typing import Type

import pandas as pd

from pandabear.model import DataFrameModel, SeriesModel


class DataFrame(pd.DataFrame, DataFrameModel):
    @classmethod
    def __class_getitem__(cls, typ: Type[DataFrameModel]):
        """Enable DataFrame[MySchema] syntax in type hints.

        Setting `DataFrame[MySchema]` to return `pd.DataFrame | MySchema` allows
        `pandabear` to play nice with other runtime type checkers such as `pydantic`
        and `beartype`. Notice how the `check_schemas` decorator handles type hints
        in case the the value is a `pd.DataFrame`. It will check if the type hint
        is a `pd.DataFrame | MySchema` and if so, it will validate the dataframe
        against `MySchema`.
        """
        return pd.DataFrame | typ


class Series(pd.Series, SeriesModel):
    @classmethod
    def __class_getitem__(cls, typ: Type[SeriesModel]):
        """Enable Series[MySeries] syntax in type hints.

        Setting e.g. `Series[MySeries]` to return `pd.Series | MySeries` allows
        `pandabear` to play nice with other runtime type checkers such as `pydantic`
        and `beartype`. Notice how the `check_schemas` decorator handles type hints
        in case the the value is a `pd.Series`. It will check if the type hint
        is a `pd.Series | MySeries` and if so, it will validate the Series
        against `MySeries`.
        """
        return pd.Series | typ



import re
from types import NoneType
from typing import Any, Union

import numpy as np
import pandas as pd

from pandabear.column_checks import CHECK_NAME_FUNCTION_MAP
from pandabear.exceptions import (
    CoersionError,
    ColumnCheckError,
    MissingColumnsError,
    MissingIndexError,
    SchemaDefinitionError,
    SchemaValidationError,
)
from pandabear.model_components import BaseConfig, Field, FieldInfo, Index

TYPE_DTYPE_MAP = {
    str: np.dtype("O"),
}


# @dataclasses.dataclass
class BaseModel:
    Config: BaseConfig = BaseConfig

    @classmethod
    def _get_config(cls):
        """Get the config for the model.

        Schema definitions that inherit from this class often overrides the
        `Config` class attribute, like e.g.:

        >>> class MySchema(DataFrameModel):
        >>>     class Config:
        >>>         coerce = True

        This is basically a complete override of `cls.Config` that removes all
        the other (default) `Config` attributes. This method is used as a getter
        to retrieve the `Config` class attribute, with all the default values
        intact.
        """
        return BaseConfig._override(cls.Config)

    @classmethod
    def _validate_series(cls, se: pd.Series, field: Field, typ: Any, coerce: bool) -> pd.Series:
        """Validate a series against a field and type.

        Args:
            se (pd.Series): The series to validate.
            field (Field): The field to validate against.
            typ (Any): The type to validate against.
            coerce (bool): Whether to coerce the series to the type of the
                field.

        Returns:
            pd.Series: The validated series.
        """
        dtype = TYPE_DTYPE_MAP.get(typ, typ)

        if se.dtype != dtype:
            if coerce:
                try:
                    se = se.astype(typ)
                except ValueError:
                    raise CoersionError(f"Could not coerce `{se.name}` with dtype {se.dtype} to {dtype}")
            else:
                raise SchemaValidationError(f"Expected `{se.name}` with dtype {dtype} but found {se.dtype}")

        for check_name, check_func in CHECK_NAME_FUNCTION_MAP.items():
            check_value = getattr(field, check_name)
            if check_value is not None:
                result = check_func(series=se, value=check_value)
                if not result.all():
                    raise ColumnCheckError(check_name=check_name, check_value=check_value, series=se, result=result)
        return se


class DataFrameModel(BaseModel):
    @classmethod
    def _get_schema_map(cls) -> dict[str, FieldInfo]:
        """Get a convenient representation of the schema.

        This method builds a dictionary that maps index/column names to a
        tuple containing (type, optional, is_index, field). This is a
        convenient representation of the schema, because it allows easy access
        to otherwise hard-to-get information about the schema.

        Note: The `getattr(cls, name) if hasattr(cls, name) else Field()` bit
            assigns a `Field` value to columns that are defined in the schema
            without a `Field`. This is useful, because it allows a more concise
            API where columns that don't have aliases or need checks can be
            defined as an annotation without a `Field` object. E.g.:
            >>> class MySchema(DataFrameModel):
            >>>     column_a: int
            >>>     column_b: str = Field(alias="column_b_alias")
            Here `column_a` has no checks and no alias, so it is defined simply
            as an annotation.

        Returns:
            schema_map (dict): A dictionary mapping index/column names to a
                tuple containing (type, optional, is_index, field)
        """
        schema_map = {}
        for name, typ in cls.__annotations__.items():
            typ, optional = cls._check_optional_type(typ)
            is_index = False
            if hasattr(typ, "__args__") and typ.__args__[0] is Index:
                # `typ` is like `Union[Index, int]` (meaning the user provided `Index[int]`)
                typ = typ.__args__[1]
                is_index = True
            schema_map[name] = (typ, optional, is_index, getattr(cls, name) if hasattr(cls, name) else Field())
        return schema_map

    @staticmethod
    def _check_optional_type(typ: type) -> tuple[type, bool]:
        """Check if a type is optional and return the non-optional type."""
        optional = False
        if hasattr(typ, "__args__") and type(None) in typ.__args__:
            optional = True
            typ = Union[tuple(arg for arg in typ.__args__ if arg is not type(None))]
        return typ, optional

    @staticmethod
    def override_level(
        index: pd.MultiIndex | pd.Index, index_level: str, series: pd.Series
    ) -> pd.MultiIndex | pd.Index:
        """Override a level in a MultiIndex or Index with a new series."""
        if isinstance(index, pd.MultiIndex):
            df_reset = index.to_frame(index=False)
            if index_level not in df_reset.columns:
                raise ValueError(f"Index level '{index_level}' not found in MultiIndex.")
            df_reset[index_level] = series.values
            new_index = pd.MultiIndex.from_frame(df_reset)
        else:
            if index.name != index_level:
                raise ValueError(f"Index name '{index.name}' does not match given index_level '{index_level}'.")
            new_index = pd.Index(series.values, name=index_level)
        return new_index

    @staticmethod
    def _select_index_series(df: pd.DataFrame, level: str, optional: bool = True) -> list[pd.Series]:
        """Select a series from a dataframe by column name.

        Return a list containing maximally 1 series. Reason for this is that
        series are validated in a loop, so returning a list is convenient.
        """
        try:
            return [df.index.get_level_values(level).to_series()]
        except KeyError:
            # When this happens we can deduce that the corresponding column is
            # optional (otherwise an error would have been raised in
            # _validate_columns).
            assert (
                optional
            ), "This should not happen. Looks like columns were not properly filtered in `_validate_columns`"
            return []

    @staticmethod
    def _select_series(df: pd.DataFrame, column_name: str, optional: bool = True) -> list[pd.Series]:
        """Select a series from a dataframe by column name.

        Return a list containing maximally 1 series. Reason for this is that
        series are validated in a loop, so returning a list is convenient.
        """
        try:
            return [df[column_name]]
        except KeyError:
            # When this happens we can deduce that the corresponding column is
            # optional (otherwise an error would have been raised in
            # _validate_columns).
            assert (
                optional
            ), "This should not happen. Looks like columns were not properly filtered in `_validate_columns`"
            return []

    @staticmethod
    def _select_index_series_by_regex(df: pd.DataFrame, alias: str) -> list[pd.Series]:
        """Select a series from a dataframe by regex."""
        return [
            df.index.get_level_values(level).to_series()
            for level in df.index.names
            if re.match(alias, level) is not None
        ]

    @staticmethod
    def _select_series_by_regex(df: pd.DataFrame, alias: str) -> list[pd.Series]:
        """Select a series from a dataframe by regex."""
        return [df[col] for col in df.filter(regex=alias, axis=1).columns]

    @classmethod
    def _validate_custom_checks(cls, df: pd.DataFrame):
        """Validate custom checks defined on the schema.

        The `check` decorator can be used to define custom checks on the
        schema. This method will loop through all attributes on the schema and
        check if they have the `__check__` attribute. If they do, it means they
        are decorated with the `check` decorator and should be run on the
        dataframe.
        """
        for attr_name in dir(cls):
            attr = getattr(cls, attr_name)
            if not hasattr(attr, "__check__"):
                continue

            check_columns: list[str] | NoneType = getattr(attr, "__check__")

            if check_columns is None:
                # assumes check is for whole df
                if not attr(df):
                    raise ValueError(f"DataFrame did not pass custom check `{attr_name}`")
                continue

            if any(undefined_columns := [c for c in check_columns if c not in cls.__annotations__]):
                raise SchemaDefinitionError(
                    f"Decorator on custom check `{attr_name}` references undefined columns {undefined_columns}. Values passed to the `check` decorator must reference columns defined in the schema."
                )

            for column in check_columns:
                if not attr(df[column]):
                    raise ValueError(f"Column `{column}` did not pass custom check `{attr_name}`")

    @classmethod
    def _validate_schema(cls, schema_map: dict[str, FieldInfo]):
        """Validate the schema map.

        This method will raise errors if there are obvious problems with the
        schema, like e.g. missing aliases when regex=True, number checks on
        non-numeric columns, etc.
        """
        for name, (typ, optional, is_index, field) in schema_map.items():
            # Check that regex is not used when alias is not defined
            if field.regex and field.alias is None:
                raise SchemaDefinitionError(
                    f"Regex is used for `{name}` in schema `{cls.__name__}`, but no alias is defined."
                )

            # Check that string checks arenot used on non-string columns
            if any(
                [
                    field.str_contains is not None,
                    field.str_endswith is not None,
                    field.str_startswith is not None,
                ]
            ) and typ not in [str]:
                raise SchemaDefinitionError(
                    f"String check is used for `{name}` in schema `{cls.__name__}`, but the field is not a string."
                )

    @classmethod
    def _validate_multiindex(
        cls, df: pd.DataFrame, schema_map: dict[str, FieldInfo], Config: BaseConfig
    ) -> pd.DataFrame:
        """Validate index levels in `df` against the schema.

        Raise approproate errors if index levels are missing or if there is an
        overlap between index levels that different schema fields match.

        This method also acts as a filter, that (depending on the config) will
        pass `df` through with coerced types, filtered index levels, ordered
        index levels or as-is.
        """
        matching_index_names_in_df = cls._select_matching_names(list(df.index.names), schema_map, match_index=True)

        if Config.filter:
            # Make sure that only the matching index levels are kept
            if df.index.names != [None] and len(matching_index_names_in_df) < len(df.index.names):
                if len(matching_index_names_in_df) == 0:
                    df = df.reset_index(drop=True, inplace=True)
                else:
                    df = df.droplevel([ind for ind in df.index.names if ind not in matching_index_names_in_df])

        if Config.multiindex_strict:
            if unexpected_indices := set(df.index.names) - set(matching_index_names_in_df) - set([None]):
                raise SchemaValidationError(
                    f"MultiIndex names {unexpected_indices} are present in `df` but not defined in schema. Use `multiindex_strict=False` to supress this error."
                )

        if Config.multiindex_ordered:
            if df.index.names != [None] and matching_index_names_in_df != list(df.index.names):
                raise SchemaValidationError(
                    "MultiIndex names in `df` are not ordered as in schema. Use `multiindex_ordered=False` to supress this error."
                )

        if Config.multiindex_sorted:
            if not (df.index.is_monotonic_increasing or df.index.is_monotonic_decreasing):
                raise SchemaValidationError(
                    "MultiIndex is not sorted. Use `multiindex_sorted=False` to supress this error."
                )

        if Config.multiindex_unique:
            if not df.index.is_unique:
                raise SchemaValidationError(
                    "MultiIndex is not unique. Use `multiindex_unique=False` to supress this error."
                )

        return df

    @classmethod
    def _validate_columns(cls, df: pd.DataFrame, schema_map: dict[str, FieldInfo], Config: BaseConfig) -> pd.DataFrame:
        """Validate column names in `df` against the schema.

        Raise approproate errors if columns are missing or if there is an
        overlap between columns that different schema fields match.

        This method also acts as a filter, that (depending on the config) will
        pass `df` through with coerced types, filtered columns, ordered columns
        or as-is.
        """
        matching_columns_in_df = cls._select_matching_names(list(df.columns), schema_map)

        # Drop columns in `df` that do not match the schema
        if Config.filter:
            ordered_columns_in_df = [col for col in df.columns if col in matching_columns_in_df]
            df = df.copy()[ordered_columns_in_df]

        # Complain about columns in `df` that are not defined in the schema
        elif Config.strict:
            if unexpected_columns := set(df.columns) - set(matching_columns_in_df):
                raise SchemaValidationError(
                    f"Columns {unexpected_columns} are present in `df` but not in schema. Use `strict=False` or `filter=True` to supress this error."
                )

        # Complain if the order of columns in `df` does not match the order in
        # which they are defined in the schema
        if Config.ordered:
            if matching_columns_in_df != list(df.columns):
                raise SchemaValidationError(
                    "Columns in `df` are not ordered as in schema. Use `ordered=False` to supress this error."
                )

        return df

    @classmethod
    def _select_matching_names(
        cls, names: list[str], schema_map: dict[str, FieldInfo], match_index: bool = False
    ) -> list[str]:
        """Select columns or index levels in `names` that match the schema.

        If a column or index level is defined in the schema it *must* be
        present in `df`. This function serves the important purpose of raising
        errors when columns in `df` seem to be *missing* when compared to the
        schema.

        Raises:
            SchemaDefinitionError: If a column or alias is not found in `df`,
                is already matched by another field.
            MissingColumnsError: If a column is defined in the schema but not
                found in `df`.
            MissingIndexError: If an index level is defined in the schema
                but not found in `df`.
        """
        MissingNameError = MissingIndexError if match_index else MissingColumnsError
        series_type = "index level" if match_index else "column"
        matching_names = []
        for series_name, (_, optional, is_index, field) in schema_map.items():
            if is_index and not match_index:
                continue
            elif not is_index and match_index:
                continue
            if field.alias is not None and field.regex:
                matched = [name for name in names if re.match(field.alias, name)]
                if len(matched) == 0 and not optional:
                    raise MissingNameError(
                        f"No {series_type}s match regex `{field.alias}` for field `{series_name}` in schema `{cls.__name__}`"
                    )
                elif len(already_matched := set(matched) & set(matching_names)) > 0:
                    raise SchemaDefinitionError(
                        f"Regex `{field.alias}` for field `{series_name}` in schema `{cls.__name__}` matched {series_type}s {already_matched} already matched by another field."
                    )
                matching_names.extend(matched)
            elif field.alias is not None and field.regex is False:
                if field.alias not in names and not optional:
                    raise MissingNameError(
                        f"No {series_type}s match alias `{field.alias}` for field `{series_name}` in schema `{cls.__name__}`."
                    )
                elif field.alias in matching_names:
                    raise SchemaDefinitionError(
                        f"Alias `{field.alias}` for field `{series_name}` in schema `{cls.__name__}` is used by another field."
                    )
                matching_names.append(field.alias)
            else:
                if series_name not in names and not optional:
                    raise MissingNameError(
                        f"No {series_type}s match {series_type} name `{series_name}` in schema `{cls.__name__}`."
                    )
                elif series_name in matching_names:
                    raise SchemaDefinitionError(
                        f"{series_type.capitalize()} `{series_name}` in schema `{cls.__name__}` is used by another field."
                    )
                matching_names.append(series_name)
        return matching_names

    @classmethod
    def validate(cls, df: pd.DataFrame) -> pd.DataFrame:
        """Validate a dataframe against the schema.

        This method also acts as a filter, that (depending on the config) will
        pass `df` through with coerced types, filtered columns, ordered columns
        or as-is.

        Args:
            df (pandas.DataFrame): The dataframe to validate.

        Returns:
            pandas.DataFrame: The validated dataframe.

        Raises:
            MissingColumnsError: If a column name/alias/regex is not in `df`.
                This happens when a column is defined in the schema but not in
                `df`. Or (more often than you'd think) because of typos ^^.
            SchemaDefinitionError: If there is a problem with the schema. This
                could e.g. happen when there is an overlap between the columns
                in `df` that different schema fields match, so that it is not
                clear which field should be used to validate a column.
            SchemaValidationError: If there is a problem with the data in `df`.
                This could e.g. happen when a column is of the wrong dtype or
                when a custom check fails.
        """
        schema_map = cls._get_schema_map()
        Config = cls._get_config()

        # Validate schema definition. Catch errors like, e.g., missing aliases
        # when regex=True, number checks on non-numeric columns, etc.
        cls._validate_schema(schema_map)

        # Check that indices and columns in `df` match schema. The only errors
        # that should be thrown here relate to schema errors or missing columns
        # in `df`. Furthermore, this method may filter, coerce or order `df`
        # depending on user-provided specifications in `Config`.
        df = cls._validate_multiindex(df, schema_map, Config)
        df = cls._validate_columns(df, schema_map, Config)

        # Validate `df` against schema. The only errors that should be raised
        # in this step are from dtype checks and `Field` checks.
        for name, (typ, optional, is_index, field) in schema_map.items():
            # Select the column (or columns) in `df` that match the field.
            # ... when index column
            if is_index:
                if field.regex and field.alias is not None:
                    matched_series = cls._select_index_series_by_regex(df, field.alias)
                else:
                    matched_series = cls._select_index_series(df, field.alias or name, optional)

            # ... when column has aliased name
            elif field.alias is not None:
                if field.regex:
                    matched_series = cls._select_series_by_regex(df, field.alias)
                else:
                    matched_series = cls._select_series(df, field.alias, optional)

            # ... when column name is attribute name (not alias)
            else:
                matched_series = cls._select_series(df, name, optional)

            # Validate the selected column(s) against the field and type.
            for series in matched_series:
                series = cls._validate_series(series, field, typ, Config.coerce or field.coerce)
                if Config.coerce or field.coerce:
                    if is_index:
                        df.index = cls.override_level(df.index, series.name, series)
                    else:
                        df[series.name] = series

        cls._validate_custom_checks(df)

        return df


class SeriesModel(BaseModel):
    @classmethod
    def _get_value_name_and_type(cls):
        return list(cls.__annotations__.items())[0]

    @classmethod
    def _get_field(cls):
        value_name, _ = cls._get_value_name_and_type()
        return getattr(cls, value_name)

    @classmethod
    def validate(cls, series: pd.Series):
        """Validate a series against the schema.

        Args:
            series (pandas.Series): The series to validate.

        Returns:
            pandas.Series: The validated series.
        """
        _, value_type = cls._get_value_name_and_type()
        field = cls._get_field()
        Config = cls._get_config()
        series = cls._validate_series(series, field, value_type, Config.coerce)
        return series



import pandas as pd
import pytest

from pandabear import DataFrameModel, Field, check
from pandabear.exceptions import SchemaDefinitionError


class TestCustomChecksSuccessDataFrame:
    def test___custom_checks__success__dataframe__1(self):
        """Test that the custom checks works for series."""

        class MySchema(DataFrameModel):
            column_a: int = Field()

            @check("column_a")
            def check_column_a(column_a: pd.Series) -> bool:
                return column_a.sum() > 0

        df = pd.DataFrame(
            dict(
                column_a=[1, 2, 3],
            )
        )

        MySchema.validate(df)

    def test___custom_checks__success__dataframe__2(self):
        """Test that the custom checks works for multiple series."""

        class MySchema(DataFrameModel):
            column_a: int = Field()
            column_c: float = Field()

            @check(["column_a", "column_c"])
            def check_column(column: pd.Series) -> bool:
                return column.sum() > 0

        df = pd.DataFrame(
            dict(
                column_a=[1, 2, 3],
                column_c=[0.1, 0.2, 0.3],
            )
        )

        MySchema.validate(df)

    def test___custom_checks__success__dataframe__3(self):
        """Test that the custom checks works for multiple series, with multiple check functions."""

        class MySchema(DataFrameModel):
            column_a: int = Field()
            column_c: float = Field()

            @check("column_a")
            def check_column_a(column_a: pd.Series) -> bool:
                return column_a.sum() > 0

            @check("column_c")
            def check_column_c(column_c: pd.Series) -> bool:
                return column_c.sum() > 0

        df = pd.DataFrame(
            dict(
                column_a=[1, 2, 3],
                column_c=[0.1, 0.2, 0.3],
            )
        )

        MySchema.validate(df)


class TestCustomChecksFailureDataFrame:
    def test___custom_checks__failure__dataframe__1(self):
        """Test that the custom checks fails when check does not pass."""
        with pytest.raises(ValueError):

            class MySchema(DataFrameModel):
                column_a: int = Field()

                @check("column_a")
                def check_column_a(column_a: pd.Series) -> bool:
                    return column_a.sum() > 0

            df = pd.DataFrame(
                dict(
                    column_a=[-1, -2, -3],
                )
            )

            MySchema.validate(df)

    def test___custom_checks__failure__dataframe__2(self):
        """Test that the custom checks fails when one of checks do not pass."""
        with pytest.raises(ValueError):

            class MySchema(DataFrameModel):
                column_a: int = Field()
                column_b: int = Field()

                @check(["column_a", "column_b"])
                def check_column_a(column: pd.Series) -> bool:
                    return column.sum() > 0

            df = pd.DataFrame(
                dict(
                    column_a=[1, 2, 3],
                    column_b=[-1, -2, -3],
                )
            )

            MySchema.validate(df)

    def test___custom_checks__failure__dataframe__3(self):
        """Test that the custom checks fails when no column name is passed to decorator."""
        with pytest.raises(TypeError):

            class MySchema(DataFrameModel):
                column_a: int = Field()
                column_b: int = Field()

                @check()
                def check_column(column: pd.Series) -> bool:
                    return column.sum() > 0

            df = pd.DataFrame(
                dict(
                    column_a=[1, 2, 3],
                    column_b=[-1, -2, -3],
                )
            )

            MySchema.validate(df)

    def test___custom_checks__failure__dataframe__4(self):
        """Test that the custom checks fails when bad column names are passed to decorator."""
        with pytest.raises(SchemaDefinitionError):

            class MySchema(DataFrameModel):
                column_a: int = Field()
                column_b: int = Field()

                @check("column_c")
                def check_column(column: pd.Series) -> bool:
                    return column.sum() > 0

            df = pd.DataFrame(
                dict(
                    column_a=[1, 2, 3],
                    column_b=[-1, -2, -3],
                )
            )

            MySchema.validate(df)



import re
from typing import Any

import pandas as pd

MAX_FAILURE_ROWS = 10


class MissingColumnsError(Exception):
    """Raise when `df` is missing columns defined in `schema`.

    Is not raised if Config defines `strict=False` or `filter=True`.
    """

    def __init__(self, message):
        # strip trailing "." and " " from message
        message = re.sub(r"[. ]+$", "", message)
        suggestion = ". Is there a typo in the schema definition? If not, the dataframe is missing columns."
        super().__init__(message + suggestion)


class MissingIndexError(Exception):
    """Raise when `df` is missing index levels defined in `schema`.

    Is not raised if Config defines `strict=False` or `filter=True`.
    """

    def __init__(self, message):
        # strip trailing "." and " " from message
        message = re.sub(r"[. ]+$", "", message)
        suggestion = ". Is there a typo in the schema definition? If not, the dataframe is missing index levels."
        super().__init__(message + suggestion)


class SchemaDefinitionError(Exception):
    """Raise when `schema` is not a valid schema definition.

    This may happen if e.g. the user has a numerical check on a column with non-numerical values.
    """

    def __init__(self, message):
        super().__init__(message)


class SchemaValidationError(Exception):
    """Raise when `df` does not match `schema`.

    This is raised if `df` has unexpected columns of column dtypes don't match.
    If Config defines `strict=False` or `filter=True` unexpected columns won't
    raise an error, though.
    """

    def __init__(self, message):
        super().__init__(message)


class CoersionError(Exception):
    """Raise when `df` cannot be coerced to `schema`.

    This is raised if `df` has a column with a dtype that the schema defined
    dtype cannot coerce. For example, if the schema defines a column as `int`
    but the column contains `str` values, coersion will fail.
    """

    def __init__(self, message):
        super().__init__(message)


class TypeHintError(Exception):
    """Raise when the function argument or return value is not of the expected type.

    This could for example happen if the user provides an int to an argument with a
    `DataFrame` or `Series` type hint. Of if the return type hint indicates multiple
    return values, but the function only returns a single value.
    """

    def __init__(self, message):
        super().__init__(message)


class ColumnCheckError(Exception):
    """Raise when a column check fails checks defined in `Field` variable.

    Report the percentage of rows that failed the check, and display the first
    few rows that failed the check.
    """

    def __init__(self, check_name: str, check_value: Any, series: pd.Series, result: pd.Series):
        self.check_name = check_name
        self.check_value = check_value
        self.series = series
        self.result = result
        super().__init__(self._get_message())

    def _get_message(self) -> str:
        fail_series = self.series[~self.result]
        total = len(self.series)
        fails = len(fail_series)
        fail_pc = int(round(100 * fails / total))
        check_name = self.check_name.replace("series_", "")
        text_msg = (
            f"Column '{self.series.name}' failed check {check_name}({self.check_value}): "
            f"{fails} of {total} ({fail_pc} %)"
        )
        fails_msg = fail_series.head(MAX_FAILURE_ROWS).to_string()
        return f"{text_msg}\n{fails_msg}"



import pandas as pd
import pytest

from pandabear.exceptions import MissingIndexError, SchemaValidationError
from pandabear.model import DataFrameModel
from pandabear.model_components import Field, Index


class IndexStrictConfig:
    multiindex_strict: bool = True


class IndexOrderedConfig:
    multiindex_ordered: bool = True


class IndexSortedConfig:
    multiindex_sorted: bool = True


class IndexUniqueConfig:
    multiindex_unique: bool = True


class NoIndexSchema(DataFrameModel):
    a: int = Field()
    b: float = Field()


class IndexSchema(DataFrameModel):
    index: Index[str]
    a: int = Field()
    b: float = Field()


def test_no_index_schema__passing():

    df = pd.DataFrame(dict(a=[1], b=[1.0]))

    # 1. passes
    NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())

    # 2. Strict
    NoIndexSchema.Config = IndexStrictConfig
    NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())

    # 3. Ordered
    NoIndexSchema.Config = IndexOrderedConfig
    NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())

    # 4. Sorted
    NoIndexSchema.Config = IndexSortedConfig
    NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())

    # 5. Unique
    NoIndexSchema.Config = IndexUniqueConfig
    NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())


def test_no_index_schema__failing():
    df = pd.DataFrame(dict(a=[1, 2], b=[1.0, 2.0]), index=pd.Index([1, 2], name="index"))
    with pytest.raises(SchemaValidationError):
        NoIndexSchema._validate_multiindex(df, NoIndexSchema._get_schema_map(), NoIndexSchema._get_config())


def test_index_schema__passing():
    df = pd.DataFrame(dict(a=[1, 2, 3], b=[1.0, 2.0, 3.0]), index=pd.Index([1, 2, 3], name="index"))

    # 1. passes
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 2. Strict
    IndexSchema.Config = IndexStrictConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 3. Ordered
    IndexSchema.Config = IndexOrderedConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 4. Sorted
    IndexSchema.Config = IndexSortedConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 5. Unique
    IndexSchema.Config = IndexUniqueConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())


def test_index_schema__failing_sorting_unique():
    df = pd.DataFrame(dict(a=[1, 2, 3, 4], b=[1.0, 2.0, 3.0, 4.0]), index=pd.Index([1, 3, 2, 3], name="index"))

    class IndexSchema(DataFrameModel):
        index: Index[str]
        a: int = Field()
        b: float = Field()

    # 1. passes
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 2. Strict
    IndexSchema.Config = IndexStrictConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 3. Ordered
    IndexSchema.Config = IndexOrderedConfig
    IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 4. Sorted fails
    with pytest.raises(SchemaValidationError):
        IndexSchema.Config = IndexSortedConfig
        IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())

    # 5. Unique, failing
    with pytest.raises(SchemaValidationError):
        IndexSchema.Config = IndexUniqueConfig
        IndexSchema._validate_multiindex(df, IndexSchema._get_schema_map(), IndexSchema._get_config())


def test_multiindex_schema__passing():
    class MultiIndexSchema(DataFrameModel):
        ix0: Index[int] = Field()
        ix1: Index[int] = Field()
        a: int = Field()

    df = pd.DataFrame(
        dict(a=[1, 2, 3, 4], b=[1.0, 2.0, 3.0, 4.0]),
        index=pd.MultiIndex.from_tuples([(1, 1), (1, 2), (2, 1), (2, 2)], names=["ix0", "ix1"]),
    )

    # 1. passes
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 2. ordered
    MultiIndexSchema.Config = IndexOrderedConfig
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 3. sorted
    MultiIndexSchema.Config = IndexSortedConfig
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 4. unique
    MultiIndexSchema.Config = IndexUniqueConfig
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())


def test_multiindex_schema__failing():
    class MultiIndexSchema(DataFrameModel):
        ix0: Index[int] = Field()
        ix1: Index[int] = Field()
        a: int = Field()

    df = pd.DataFrame(
        dict(a=[1, 2, 3, 4], b=[1.0, 2.0, 3.0, 4.0]),
        index=pd.MultiIndex.from_tuples([(1, 1), (1, 1), (2, 2), (2, 1)], names=["ix0", "ix1"]),
    )

    # 1. passes
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 3. sorted
    with pytest.raises(SchemaValidationError):
        MultiIndexSchema.Config = IndexSortedConfig
        MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 4. unique
    with pytest.raises(SchemaValidationError):
        MultiIndexSchema.Config = IndexUniqueConfig
        MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 2. ordered, passing
    MultiIndexSchema.Config = IndexOrderedConfig
    MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    # 2. ordered, failing
    with pytest.raises(SchemaValidationError):
        MultiIndexSchema.Config = IndexOrderedConfig
        df.index.names = ["ix1", "ix0"]
        MultiIndexSchema._validate_multiindex(df, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())

    df2 = df.reset_index()
    with pytest.raises(MissingIndexError):
        MultiIndexSchema._validate_multiindex(df2, MultiIndexSchema._get_schema_map(), MultiIndexSchema._get_config())



import pandas as pd
import pytest

from pandabear import DataFrameModel, Field
from pandabear.decorators import dataframe_check


def test__validate_custom_dataframe_check__success():
    """Test that the custom checks works for dataframes."""

    class MySchema(DataFrameModel):
        column_a: int = Field()
        column_b: int = Field()

        @dataframe_check
        def check_df(df: pd.DataFrame) -> bool:
            return df.sum().sum() > 0

    df = pd.DataFrame(
        dict(
            column_a=[1, 2, 3],
            column_b=[1, 0, 3],
        )
    )

    assert hasattr(MySchema.check_df, "__check__")
    assert getattr(MySchema.check_df, "__check__") is None

    MySchema.validate(df)


def test__validate_custom_dataframe_check__failure():
    """Test that the custom checks fails foir datafrane check."""

    class MySchema(DataFrameModel):
        column_a: int = Field()
        column_b: int = Field()

        @dataframe_check
        def check_df(df: pd.DataFrame) -> bool:
            return (df > 0).all().all()

    df = pd.DataFrame(
        dict(
            column_a=[1, 2, 3],
            column_b=[-1, 1, 2],
        )
    )

    with pytest.raises(ValueError):
        MySchema.validate(df)



import pandas as pd
import pytest

from pandabear.exceptions import MissingColumnsError, SchemaValidationError
from pandabear.model import DataFrameModel, SeriesModel
from pandabear.model_components import Field


def test_strict_filter_ordered_columns():
    class MySchema(DataFrameModel):
        a: int = Field()
        b: float = Field()
        c: str = Field()

    # 1. passes, with a,b,c and not d present.
    class FilterConfig:
        filter: bool = True

    MySchema.Config = FilterConfig
    df = pd.DataFrame(dict(a=[1], b=[1.0], c=["a"], d=[1]))
    dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())
    assert dfval.shape == (1, 3)
    assert dfval.columns.tolist() == ["a", "b", "c"]

    # 2. column order is maintained
    df = pd.DataFrame(dict(b=[1.0], a=[1], d=[1], c=["a"]))
    dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())
    assert dfval.shape == (1, 3)
    assert dfval.columns.tolist() == ["b", "a", "c"]

    # 3. changing to strict == true doesn't do anything, filter takes precedence
    class FilterConfig:
        filter: bool = True
        strict: bool = True

    MySchema.Config = FilterConfig
    df = pd.DataFrame(dict(b=[1.0], a=[1], d=[1], c=["a"]))
    dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())
    assert dfval.columns.tolist() == ["b", "a", "c"]

    # 4. Changing filter to false, strict will now take effect, and fail:
    class FilterConfig:
        filter: bool = False
        strict: bool = True

    MySchema.Config = FilterConfig
    df = pd.DataFrame(dict(b=[1.0], a=[1], d=[1], c=["a"]))
    with pytest.raises(SchemaValidationError):
        dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())

    # 5. changing to ordered == true, will now fail
    class FilterConfig:
        filter: bool = False
        strict: bool = True
        ordered: bool = True

    MySchema.Config = FilterConfig
    df = pd.DataFrame(dict(b=[1.0], a=[1], c=["a"]))
    with pytest.raises(SchemaValidationError):
        dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())

    # 6. changing to filter == true, will still fail
    class FilterConfig:
        filter: bool = True
        strict: bool = True
        ordered: bool = True

    MySchema.Config = FilterConfig
    df = pd.DataFrame(dict(b=[1.0], a=[1], c=["a"]))
    with pytest.raises(SchemaValidationError):
        dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())

    # 7. missing column will fail on filter
    df = pd.DataFrame(dict(b=[1.0], a=[1]))
    with pytest.raises(MissingColumnsError):
        dfval = MySchema._validate_columns(df, MySchema._get_schema_map(), MySchema._get_config())



import dataclasses
from typing import Any, NamedTuple, Type


@dataclasses.dataclass
class Field:
    """Class for defining a schema for a column (or group thereof) in a dataframe.

    Args:
        typ: The type of the column. If None, the type is not checked.
        ge: Greater than or equal to.
        gt: Greater than.
        le: Less than or equal to.
        lt: Less than.
        isin: Checks if the column is in the given list.
        notin: Checks if the column is not in the given list.
        str_contains: Checks if the column contains the given string.
        str_startswith: Checks if the column starts with the given string.
        str_endswith: Checks if the column ends with the given string.
        notnull: Checks if the column is not null.
        null: Checks if the column is null.
        unique: Checks if the column is unique.
        alias: Alias for the column name. Can be a regex.
        regex: Whether or not the alias is a regex.
        coerce: Whether or not to coerce the column to the given type.
    """

    typ: Any = None

    # Pandas Series checks
    ge: int | float | None = None
    gt: int | float | None = None
    lt: int | float | None = None
    le: int | float | None = None
    isin: list | None = None
    notin: list | None = None
    str_contains: str | None = None
    str_endswith: str | None = None
    str_startswith: str | None = None
    notnull: bool | None = None
    null: bool | None = None
    unique: bool | None = None

    # Column name checks
    alias: str | None = None
    regex: bool = False
    coerce: bool = False


@dataclasses.dataclass
class BaseConfig:
    strict: bool = True
    filter: bool = False
    ordered: bool = False
    coerce: bool = False
    multiindex_strict: bool = True
    multiindex_ordered: bool = False
    multiindex_sorted: bool = False
    multiindex_unique: bool = False

    @classmethod
    def _override(cls, other_cls):
        if other_cls is cls:
            return cls
        cls._assert_config_fields(other_cls)
        cls._assert_config_types(other_cls)
        new_class = type("SchemaConfig", (other_cls, cls), {})
        new_class.__annotations__ = cls.__annotations__.copy()
        return new_class

    @classmethod
    def _assert_config_fields(cls, other_cls):
        for name in dir(other_cls):
            if name.startswith("_"):
                continue
            if name in cls.__annotations__:
                continue
            raise ValueError(f"Config field `{name}` is not defined in BaseConfig")

    @classmethod
    def _assert_config_types(cls, other_cls):
        annotations = cls.__annotations__
        for name in dir(other_cls):
            if name.startswith("_"):
                continue
            if name in cls.__annotations__:
                value = getattr(other_cls, name)
                expeced_typ = annotations[name]
                if not isinstance(value, expeced_typ):
                    raise TypeError(f"Config field `{name}` expected type {expeced_typ} but found {type(value)}")


class Index:
    @classmethod
    def __class_getitem__(cls, typ):
        """Only for "marking" index columns as part of index."""
        return cls | typ


class FieldInfo(NamedTuple):
    type: Type
    optional: bool
    is_index: bool
    field: Field



import pytest

from pandabear.model import DataFrameModel, Field
from pandabear.model_components import BaseConfig


def test_default_config():
    """This test checks whether the default configuration values of a
    DataFrameModel subclass are the same as the default configuration values
    of the BaseConfig class."""

    class MySchema(DataFrameModel):
        a: int
        b: str

    Config = MySchema._get_config()
    for name, typ in Config.__annotations__.items():
        val = getattr(Config, name)
        print(name, typ, val)
        assert getattr(Config, name) == getattr(BaseConfig, name)
        assert isinstance(typ, type)


def test_override_config():
    """This test checks whether the _override method of the BaseConfig class
    works correctly with DataFrameModel subclasses that have inheritance.
    """

    # 1. Test of BaseConfig override works

    class MyConfig:
        filter = True

    Config = BaseConfig._override(MyConfig)

    print(Config.__annotations__)
    assert Config.__annotations__ == BaseConfig.__annotations__
    assert Config.filter is True
    assert Config.strict is True

    # 2. Test if this works with DataFrameModel

    class MySchema(DataFrameModel):
        a: int
        b: str
        Config = MyConfig

    Config = MySchema._get_config()

    assert Config.__annotations__ == BaseConfig.__annotations__
    assert Config.filter is True

    # 3. test that it works for DataFrameModel with inheritance

    class MySchema(DataFrameModel):
        a: int
        b: str
        Config = MyConfig

    class MySchema2(MySchema):
        c: float

    Config = MySchema2._get_config()

    assert Config.__annotations__ == BaseConfig.__annotations__
    assert Config.filter is True


def test_assert_config_fields():
    class BadConfig:
        BAD_FIELD = True

    expected_msg = "Config field `BAD_FIELD` is not defined in BaseConfig"
    with pytest.raises(ValueError, match=expected_msg):
        BaseConfig._assert_config_fields(BadConfig)


def test_assert_config_types():
    class BadConfig:
        filter = 1

    expected_msg = "Config field `filter` expected type <class 'bool'> but found <class 'int'>"
    with pytest.raises(TypeError, match=expected_msg):
        BaseConfig._assert_config_types(BadConfig)


if __name__ == "__main__":
    test_default_config()
    test_override_config()
    print("Done")



"""Test the base case for the `check_schemas` decorator.

If these tests pass, we can expect that the `check_schemas` decorator will work
in cases such as (and with great variation of):

```
    @check_schemas
    def my_function(df: DataFrame[MySchema]) -> Series[MySeries]:
        return df.column_a
    
    column_a = my_function(df)
```

When the `check_schemas` decorator is applied to `my_function`, and the function
call does not raise an error, we are guaranteed that `df` is a `DataFrame` that
follows the `MySchema` schema, and that the return value `column_a` is a `Series`
that follows the `MySeries` type definition. Neat right?!
"""

import pandas as pd
import pytest

from pandabear import (
    DataFrame,
    DataFrameModel,
    Field,
    Series,
    SeriesModel,
    check_schemas,
)
from pandabear.exceptions import ColumnCheckError, TypeHintError


# Define a custom dataframe schema
class MySchema(DataFrameModel):
    column_a: int = Field(gt=0)
    column_b: str = Field(str_contains="foo")
    column_c: float = Field(ge=0.0, le=1.0)
    my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

    class Config:
        strict = False


class MySchemaFailure(DataFrameModel):
    column_a: int = Field(gt=1)
    column_b: str = Field(str_contains="bar")
    column_c: float = Field(ge=0.0, le=0.1)
    my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=5)


# Define a custom series schema
class MySeries(SeriesModel):
    value: int = Field(gt=0, lt=10)


class MySeriesFailure(SeriesModel):
    value: int = Field(gt=2, lt=2)


# Define an example dataframe that follows the dataframe schema
df = pd.DataFrame(
    dict(
        column_a=[1, 2, 3],
        column_b=["foo", "foo-bar", "baz-foo"],
        column_c=[0.1, 0.2, 0.3],
        my_prefix_column_a=[1, 2, 3],
        my_prefix_column_b=[4, 5, 6],
    )
)

se = df.column_a


class TestBaseCasesSuccessDataFrame:
    def test___base_case__success__dataframe__1(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(df: DataFrame[MySchema]) -> DataFrame[MySchema]:
            return df

        my_function(df)

    def test___base_case__success__dataframe__multiple_in(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(dfs: tuple[DataFrame[MySchema], DataFrame[MySchema]]) -> DataFrame[MySchema]:
            df, df1 = dfs
            return df

        my_function([df, df])

    def test___base_case__success__dataframe__multiple_out(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(df: DataFrame[MySchema]) -> tuple[DataFrame[MySchema], DataFrame[MySchema]]:
            return df, df

        my_function(df)

    def test___base_case__success__dataframe__nested_in__1(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(
            nested_df: tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]
        ) -> DataFrame[MySchema]:
            df, (df1, df2) = nested_df
            return df

        my_function([df, [df, df]])

    def test___base_case__success__dataframe__nested_in__2(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(
            nested_df: tuple[
                DataFrame[MySchema], tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]
            ]
        ) -> DataFrame[MySchema]:
            df, (df1, (df2, df3)) = nested_df
            return df

        my_function([df, [df, [df, df]]])

    def test___base_case__success__dataframe__nested_out__1(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(
            df: DataFrame[MySchema],
        ) -> tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]:
            return df, [df, df]

        my_function(df)

    def test___base_case__success__dataframe__nested_out__2(self):
        """Test that the base case works for dataframes."""

        @check_schemas
        def my_function(
            df: DataFrame[MySchema],
        ) -> tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]]:
            return df, [df, [df, df]]

        my_function(df)

    def test___base_case__success__dataframe__with_beartype(self):
        """Test that the base case works for dataframes with beartype."""
        from beartype import beartype

        @beartype
        @check_schemas
        def my_function(df: DataFrame[MySchema], val: int) -> tuple[DataFrame[MySchema], int]:
            return df, val

        my_function(df, 1)

    def test___base_case__success__dataframe__filtering__1(self):
        """Test that filtering propagages filtered data in/out of function."""

        class MySchemaMissingColumnC(DataFrameModel):
            column_a: int = Field(gt=0)
            column_b: str = Field(str_contains="foo")
            my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

            class Config:
                filter = True

        @check_schemas
        def my_function(df: DataFrame[MySchemaMissingColumnC]) -> DataFrame[MySchemaMissingColumnC]:
            assert "column_c" not in df.columns
            return df

        assert "column_c" not in my_function(df).columns

    def test___base_case__success__dataframe__filtering__2(self):
        """Test that filtering recursively propagages filtered data in/out of function."""

        class MySchemaMissingColumnC(DataFrameModel):
            column_a: int = Field(gt=0)
            column_b: str = Field(str_contains="foo")
            my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

            class Config:
                filter = True

        @check_schemas
        def my_function(
            dfs: tuple[
                DataFrame[MySchemaMissingColumnC],
                tuple[DataFrame[MySchemaMissingColumnC], DataFrame[MySchemaMissingColumnC]],
            ]
        ) -> tuple[
            DataFrame[MySchemaMissingColumnC],
            tuple[DataFrame[MySchemaMissingColumnC], DataFrame[MySchemaMissingColumnC]],
        ]:
            assert "column_c" not in dfs[0].columns
            assert "column_c" not in dfs[1][0].columns
            assert "column_c" not in dfs[1][1].columns
            return dfs

        dfs_out = my_function((df, (df, df)))

        assert "column_c" not in dfs_out[0].columns
        assert "column_c" not in dfs_out[1][0].columns
        assert "column_c" not in dfs_out[1][1].columns

    def test___base_case__success__dataframe__filtering__3(self):
        """Test that filtering plays nice with non-DataFrame arguments."""

        class MySchemaMissingColumnC(DataFrameModel):
            column_a: int = Field(gt=0)
            column_b: str = Field(str_contains="foo")
            my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

            class Config:
                filter = True

        @check_schemas
        def my_function(
            df: DataFrame[MySchemaMissingColumnC], value: int
        ) -> tuple[DataFrame[MySchemaMissingColumnC], int]:
            assert "column_c" not in df.columns
            assert value == 1
            return df, value

        output = my_function(df, 1)
        assert "column_c" not in output[0].columns
        assert output[1] == 1

    def test___base_case__success__dataframe__filtering__4(self):
        """Test that filtering plays nice with non-DataFrame unspecified arguments."""

        class MySchemaMissingColumnC(DataFrameModel):
            column_a: int = Field(gt=0)
            column_b: str = Field(str_contains="foo")
            my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

            class Config:
                filter = True

        @check_schemas
        def my_function(df: DataFrame[MySchemaMissingColumnC], *args):
            assert "column_c" not in df.columns
            assert args == ("lol",)

        my_function(df, "lol")

    def test___base_case__success__dataframe__filtering__5(self):
        """Test that filtering plays nice with non-DataFrame keyword arguments."""

        class MySchemaMissingColumnC(DataFrameModel):
            column_a: int = Field(gt=0)
            column_b: str = Field(str_contains="foo")
            my_prefix_column: int = Field(alias="my_prefix.+", regex=True, ge=0)

            class Config:
                filter = True

        @check_schemas
        def my_function(df: DataFrame[MySchemaMissingColumnC], **kwargs):
            assert "column_c" not in df.columns
            assert kwargs == {"lol": 1}

        my_function(df, lol=1)


class TestBaseCasesSuccessSeries:
    def test___base_case__success__series(self):
        """Test that the base case works for series."""

        @check_schemas
        def my_function(se: Series[MySeries]) -> Series[MySeries]:
            return se

        my_function(se)

    def test___base_case__success__series__multiple_out(self):
        """Test that the base case works for series."""

        @check_schemas
        def my_function(se: Series[MySeries]) -> tuple[Series[MySeries], Series[MySeries]]:
            return se, se

        my_function(se)

    def test___base_case__success__series__nested_out__1(self):
        """Test that the base case works for series."""

        @check_schemas
        def my_function(
            se: Series[MySeries],
        ) -> tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeries]]]:
            return se, (se, se)

        my_function(se)

    def test___base_case__success__series__nested_out__2(self):
        """Test that the base case works for series."""

        @check_schemas
        def my_function(
            se: Series[MySeries],
        ) -> tuple[Series[MySeries], tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeries]]]]:
            return se, (se, (se, se))

        my_function(se)

    def test___base_case__success__series__with_beartype(self):
        """Test that the base case works for series with beartype."""
        from beartype import beartype

        @beartype
        @check_schemas
        def my_function(se: Series[MySeries], val: int) -> tuple[Series[MySeries], int]:
            return se, 1

        my_function(se, 1)


class TestBaseCaseFailureDataFrame:
    def test___base_case__failure__dataframe__in(self):
        """Test that the base case fails for dataframes, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(df: DataFrame[MySchemaFailure]) -> DataFrame[MySchema]:
                return df

            my_function(df)

    def test___base_case__failure__dataframe__out(self):
        """Test that the base case fails for dataframes, when output fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(df: DataFrame[MySchema]) -> DataFrame[MySchemaFailure]:
                return df

            my_function(df)

    def test___base_case__failure__dataframe__nested__in(self):
        """Test that the base case fails for dataframes, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(
                df: tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchemaFailure]]]
            ) -> DataFrame[MySchema]:
                return df

            my_function([df, [df, df]])

    def test___base_case__failure__dataframe__nested__out(self):
        """Test that the base case fails for dataframes, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(
                df: DataFrame[MySchema],
            ) -> tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchemaFailure]]]:
                return [df, [df, df]]

            my_function(df)

    def test___base_case__failure__dataframe__mismatch_in__1(self):
        """Test that the base case fails for dataframes, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(tup: tuple[DataFrame[MySchema], DataFrame[MySchema]]) -> DataFrame[MySchema]:
                return tup[0]

            my_function(df)

    def test___base_case__failure__dataframe__mismatch_in__2(self):
        """Test that the base case fails for dataframes, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(df: DataFrame[MySchema]) -> DataFrame[MySchema]:
                return df

            my_function((df, df))

    def test___base_case__failure__dataframe__mismatch_in__3(self):
        """Test that the base case fails for dataframes, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(tup: tuple[DataFrame[MySchema], DataFrame[MySchema]]) -> DataFrame[MySchema]:
                return tup[0]

            my_function((df, [df, df]))

    def test___base_case__failure__dataframe__mismatch_in__4(self):
        """Test that the base case fails for dataframes, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(
                nested_df: tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]
            ) -> DataFrame[MySchema]:
                df, (df1, df2) = nested_df
                return df

            my_function((df, df, df))

    def test___base_case__failure__dataframe__mismatch_out__1(self):
        """Test that the base case fails for dataframes, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(df: DataFrame[MySchema]) -> tuple[DataFrame[MySchema], DataFrame[MySchema]]:
                return df

            my_function(df)

    def test___base_case__failure__dataframe__mismatch_out__2(self):
        """Test that the base case fails for dataframes, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(df: DataFrame[MySchema]) -> DataFrame[MySchema]:
                return df, df

            my_function(df)

    def test___base_case__failure__dataframe__mismatch_out__3(self):
        """Test that the base case fails for dataframes, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(df: DataFrame[MySchema]) -> tuple[DataFrame[MySchema], DataFrame[MySchema]]:
                return df, [df, df]

            my_function(df)

    def test___base_case__failure__dataframe__mismatch_out__4(self):
        """Test that the base case fails for dataframes, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(
                df: DataFrame[MySchema],
            ) -> tuple[DataFrame[MySchema], tuple[DataFrame[MySchema], DataFrame[MySchema]]]:
                return df, df, df

            my_function(df)


class TestBaseCaseFailureSeries:
    def test___base_case__failure__series__in(self):
        """Test that the base case fails for series, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(se: Series[MySeriesFailure]) -> Series[MySeries]:
                return se

            my_function(se)

    def test___base_case__failure__series__out(self):
        """Test that the base case fails for series, when output fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(se: Series[MySeries]) -> Series[MySeriesFailure]:
                return se

            my_function(se)

    def test___base_case__failure__series__nested__in(self):
        """Test that the base case fails for series, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(
                se: tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeriesFailure]]]
            ) -> Series[MySeries]:
                return se

            my_function([se, [se, se]])

    def test___base_case__failure__series__nested__out(self):
        """Test that the base case fails for series, when input is fails."""
        with pytest.raises(ColumnCheckError):

            @check_schemas
            def my_function(
                se: Series[MySeries],
            ) -> tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeriesFailure]]]:
                return [se, [se, se]]

            my_function(se)

    def test___base_case__failure__series__mismatch_in__1(self):
        """Test that the base case fails for series, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(tup: tuple[Series[MySeries], Series[MySeries]]) -> Series[MySeries]:
                return tup[0]

            my_function(se)

    def test___base_case__failure__series__mismatch_in__2(self):
        """Test that the base case fails for series, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(se: Series[MySeries]) -> Series[MySeries]:
                return se

            my_function((se, se))

    def test___base_case__failure__series__mismatch_in__3(self):
        """Test that the base case fails for series, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(tup: tuple[Series[MySeries], Series[MySeries]]) -> Series[MySeries]:
                return tup[0]

            my_function((se, [se, se]))

    def test___base_case__failure__series__mismatch_in__4(self):
        """Test that the base case fails for series, when input type hints do not match argument values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(
                nested_se: tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeries]]]
            ) -> Series[MySeries]:
                se, (se1, se2) = nested_se
                return se

            my_function((se, se, se))

    def test___base_case__failure__series__mismatch_out__1(self):
        """Test that the base case fails for series, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(se: Series[MySeries]) -> tuple[Series[MySeries], Series[MySeries]]:
                return se

            my_function(se)

    def test___base_case__failure__series__mismatch_out__2(self):
        """Test that the base case fails for series, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(se: Series[MySeries]) -> Series[MySeries]:
                return se, se

            my_function(se)

    def test___base_case__failure__series__mismatch_out__3(self):
        """Test that the base case fails for series, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(se: Series[MySeries]) -> tuple[Series[MySeries], Series[MySeries]]:
                return se, [se, se]

            my_function(se)

    def test___base_case__failure__series__mismatch_out__4(self):
        """Test that the base case fails for series, when output type hints do not match return values."""
        with pytest.raises(TypeHintError):

            @check_schemas
            def my_function(
                se: Series[MySeries],
            ) -> tuple[Series[MySeries], tuple[Series[MySeries], Series[MySeries]]]:
                return se, se, se

            my_function(se)



import pandas as pd
import pytest

from pandabear.exceptions import CoersionError, SchemaValidationError
from pandabear.model import TYPE_DTYPE_MAP, DataFrameModel
from pandabear.model_components import Field, Index


def test_coerce_base_case():
    class CoerceConfig:
        coerce: bool = True

    class MySchema(DataFrameModel):
        a: int = Field()
        b: float = Field()
        c: str = Field()

    # 1. will raise error

    df = pd.DataFrame(dict(a=["1"], b=[1], c=[2]))
    print(df.dtypes)
    expected_message = "Expected `a` with dtype <class 'int'> but found object"
    with pytest.raises(SchemaValidationError, match=expected_message):
        dfval = MySchema.validate(df)

    # 2. will coerce dtypes
    MySchema.Config = CoerceConfig
    dfval = MySchema.validate(df)
    assert dfval.dtypes.tolist() == [int, float, TYPE_DTYPE_MAP[str]]
    assert dfval.a.tolist() == [1]
    assert dfval.b.tolist() == [1.0]
    assert dfval.c.tolist() == ["2"]


class TestCoerceSuccess:
    def test___coerce__index__success__1(self):
        class MySchema(DataFrameModel):
            index: Index[str] = Field(alias="index\d", regex=True)
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=[4, 5, 6]), index=[1, 2, 3]).rename_axis("index0")

        df_out = MySchema.validate(df)

        assert df_out.index.dtype == "object"

    def test___coerce__index__success__2(self):
        class MySchema(DataFrameModel):
            index: Index[int] = Field(alias="index\d", regex=True)
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=[4, 5, 6]), index=["1", "2", "3"]).rename_axis("index0")

        df_out = MySchema.validate(df)

        assert df_out.index.dtype == int

    def test___coerce__multi_index__success__1(self):
        class MySchema(DataFrameModel):
            index: Index[bool] = Field(alias="index\d", regex=True)
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(
            data=dict(column_a=[1, 2, 3]),
            index=pd.MultiIndex.from_arrays([["foo", "bar", "foo"], [0, 1, 2]], names=["index0", "index1"]),
        )

        df_out = MySchema.validate(df)

        assert df_out.index.get_level_values("index0").dtype == bool
        assert df_out.index.get_level_values("index1").dtype == bool

    def test___coerce__multi_index__success__2(self):
        class MySchema(DataFrameModel):
            index: Index[bool] = Field(alias="index\d", regex=True, coerce=True)
            column_a: int = Field(ge=0)

        df = pd.DataFrame(
            data=dict(column_a=[1, 2, 3]),
            index=pd.MultiIndex.from_arrays([["foo", "bar", "foo"], [0, 1, 2]], names=["index0", "index1"]),
        )

        df_out = MySchema.validate(df)

        assert df_out.index.get_level_values("index0").dtype == bool
        assert df_out.index.get_level_values("index1").dtype == bool

    def test___coerce__column__success__1(self):
        class MySchema(DataFrameModel):
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=["4", "5", "6"]))

        df_out = MySchema.validate(df)

        assert df_out["column_a"].dtype == int

    def test___coerce__column__success__2(self):
        class MySchema(DataFrameModel):
            column_a: str = Field()

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=[4, 5, 6]))

        df_out = MySchema.validate(df)

        assert df_out["column_a"].dtype == "object"


class TestCoerceFailure:
    def test___coerce__index__failure(self):
        class MySchema(DataFrameModel):
            index: Index[int] = Field(alias="index\d", regex=True)
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=[4, 5, 6]), index=["a", "b", "c"]).rename_axis("index0")

        with pytest.raises(CoersionError):
            MySchema.validate(df)

    def test___coerce__column__failure(self):
        class MySchema(DataFrameModel):
            column_a: int = Field(ge=0)

            class Config:
                coerce = True

        df = pd.DataFrame(dict(column_a=["a", "b", "c"]))

        with pytest.raises(CoersionError):
            MySchema.validate(df)



"""Test that Optional fields work as expected.

The user should be able to provide a field that is optional. For example:

    class MySchema(DataFrameModel):
        column_a: Optional[int] = Field()

        class Config:
            strict = True

    df = pd.DataFrame({})
    MySchema.validate(df)

Should pass without any errors.
"""

from typing import Optional

import pandas as pd
import pytest

from pandabear import DataFrame, DataFrameModel, Field
from pandabear.exceptions import MissingColumnsError
from pandabear.model_components import Index


class TestOptionalSuccess:
    def test___optional__success__column_name(self):
        """Test that optional fields work in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: Optional[int] = Field()
            column_b: int = Field()

        MySchema.validate(
            pd.DataFrame(
                dict(
                    column_a=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )

        MySchema.validate(
            pd.DataFrame(
                dict(
                    # column_a=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )

    def test___optional__success__index(self):
        """Test that optional fields work in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: Optional[Index[int]] = Field()
            column_b: int = Field()

        MySchema.validate(
            df=pd.DataFrame(dict(column_b=[4, 5, 6]), index=[1, 2, 3], columns=["column_b"]).rename_axis("column_a")
        )

        MySchema.validate(
            pd.DataFrame(
                dict(
                    column_b=[4, 5, 6],
                )
            )
        )

    def test___optional__success__alias(self):
        """Test that optional fields work in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: Optional[int] = Field(alias="column_x")
            column_b: int = Field()

        MySchema.validate(
            pd.DataFrame(
                dict(
                    column_x=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )

        MySchema.validate(
            pd.DataFrame(
                dict(
                    # column_x=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )

    def test___optional__success__regex(self):
        """Test that optional fields work in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: Optional[int] = Field(alias="series.+", regex=True)
            column_b: int = Field()

        MySchema.validate(
            pd.DataFrame(
                dict(
                    series_99=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )

        MySchema.validate(
            pd.DataFrame(
                dict(
                    # series_99=[1, 2, 3],
                    column_b=[4, 5, 6],
                )
            )
        )


class TestOptionalFailure:
    def test___optional__failure__column_name(self):
        """Test that optional fields fails in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: int = Field()
            column_b: int = Field()

        with pytest.raises(MissingColumnsError):
            MySchema.validate(
                pd.DataFrame(
                    dict(
                        # column_a=[1, 2, 3],
                        column_b=[4, 5, 6],
                    )
                )
            )

    def test___optional__failure__alias(self):
        """Test that optional fields fails in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: int = Field(alias="column_x")
            column_b: int = Field()

        with pytest.raises(MissingColumnsError):
            MySchema.validate(
                pd.DataFrame(
                    dict(
                        # column_x=[1, 2, 3],
                        column_b=[4, 5, 6],
                    )
                )
            )

    def test___optional__failure__regex(self):
        """Test that optional fields fails in the simplest case."""

        class MySchema(DataFrameModel):
            column_a: int = Field(alias="series.+", regex=True)
            column_b: int = Field()

        with pytest.raises(MissingColumnsError):
            MySchema.validate(
                pd.DataFrame(
                    dict(
                        # series_99=[1, 2, 3],
                        column_b=[4, 5, 6],
                    )
                )
            )



import datetime

import numpy as np
import pandas as pd


def check_dtype_equality(series_or_index, typ):
    return series_or_index.dtype == typ


def check_isinstance(series_or_index, typ):
    return isinstance(series_or_index, typ)


def check_str_object(series_or_index, typ):
    return check_dtype_equality(series_or_index, np.dtype("O"))


def check_datetime64(series_or_index, typ):
    return check_dtype_equality(series_or_index, np.dtype("datetime64[ns]"))


def check_str_expensive(series_or_index, typ):
    if not check_str_object(series_or_index, None):
        return False
    return all(isinstance(x, str) for x in series_or_index)


def check_bare_categorical_dtype(series_or_index, typ):
    return check_dtype_equality(series_or_index, "category")


TYPE_CHECK_MAP = {
    int: check_dtype_equality,
    float: check_dtype_equality,
    bool: check_dtype_equality,
    str: check_str_object,
    np.datetime64: check_datetime64,
    datetime.datetime: check_datetime64,
    pd.CategoricalDtype: check_bare_categorical_dtype,
    pd.DatetimeIndex: check_isinstance,
}


def is_of_type(series_or_index, typ):
    if typ in TYPE_CHECK_MAP:
        return TYPE_CHECK_MAP[typ](series_or_index, typ)
    if isinstance(typ, type(pd.CategoricalDtype())):
        return check_dtype_equality(series_or_index, typ)
    check_dtype_equality(series_or_index, typ)



import inspect
from functools import wraps
from types import NoneType, UnionType
from typing import Any, Callable, get_args

import pandas as pd

from pandabear.exceptions import TypeHintError
from pandabear.model import BaseModel


def _validate_variable_against_type_hint(var: Any, type_hint: Any, name: str) -> Any:
    """Validate a variable against a type hint.

    This function is used by the `check_schemas` decorator to validate
    input arguments and return values of a function.

    Args:
        var (Any): The variable to validate.
        type_hint (Any): The type hint to validate against.

    Raises:
        TypeError: If the variable does not match the type hint.
    """
    # type hint like: `pd.DataFrame | MySchema`
    if (
        isinstance(type_hint, UnionType)
        and len(get_args(type_hint)) == 2
        and issubclass(get_args(type_hint)[1], BaseModel)
    ):
        if not type(var) in [pd.DataFrame, pd.Series]:
            expected_type = get_args(type_hint)[0].__name__
            expected_schema = type_hint.__args__[1].__name__
            raise TypeHintError(
                f"Expected `{expected_type}[{expected_schema}]` in {f'argument `{name}`' if name != 'return value' else name}, but found {type(var)}"
            )
        schema = get_args(type_hint)[1]
        transformed_var = schema.validate(var)

    # type hint like: `tuple[int, pd.DataFrame | MySchema]` (or deeper nesting)
    elif (len(return_types := get_args(type_hint))) > 1:
        if type(var) not in [list, tuple]:
            raise TypeHintError(
                f"Expected a `tuple` or `list` in {f'argument `{name}`' if name != 'return value' else name}, but found {type(var)}"
            )
        elif len(var) != len(return_types):
            raise TypeHintError(
                f"Expected iterable of {len(return_types)} items in {f'argument `{name}`' if name != 'return value' else name}, but found {len(var)}"
            )
        transformed_var = ()
        for var_i, type_hint_i in zip(var, return_types):
            transformed_var += (_validate_variable_against_type_hint(var_i, type_hint_i, name),)

    # type hint is not a `DataFrameModel` subclass
    else:
        transformed_var = var

    return transformed_var


def check_schemas(func: Callable[..., Any]) -> Callable[..., Any]:
    """Main decorator for validating schemas of input and return dataframes.

    This decorator is used to validate the dataframe schema of input arguments
    and return values of a function. It will check if those input/return values
    whose type hints are `DataFrameModel` subclasses respect the schema of those
    subclasses.

    Args:
        func (Callable[..., Any]): The function to decorate.

    Returns:
        Callable[..., Any]: The decorated function.

    Raises:
        TypeError: If the input arguments or return values do not match their
            type hints.

    Examples:
        >>> from pandabear import DataFrameModel, Field, check_schemas
        >>>
        >>> class MySchema(DataFrameModel):
        >>>     column_a: int = Field()
        >>>
        >>> @check_schemas
        >>> def my_func(df: MySchema) -> MySchema:
        >>>     return df
        >>>
        >>> my_func(pd.DataFrame(dict(column_a=[1, 2, 3])))
        >>> my_func(pd.DataFrame(dict(column_a=["a", "b", "c"])))
        TypeError: Expected a pandas dataframe or series, but found <class 'str'>. Check that your type hints and returned values match.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        # Validate input argument(s)
        sig = inspect.signature(func)
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        for name, variable in bound_args.arguments.items():
            type_hint = sig.parameters[name].annotation
            bound_args.arguments[name] = _validate_variable_against_type_hint(variable, type_hint, name)

        # Extract `args` and `kwargs` from bound arguments
        args = bound_args.arguments.pop("args", {})
        kwargs = bound_args.arguments.pop("kwargs", {})

        # Execute the function
        # result = func(*args, **kwargs)
        result = func(*bound_args.arguments.values(), *args, **kwargs)

        # Validate return value(s)
        result = _validate_variable_against_type_hint(result, sig.return_annotation, "return value")

        return result

    return wrapper


def check(column_names: str | list[str]) -> Callable:
    """Decorator for defining custom checks on dataframe columns.

    This decorator is used to define custom checks on dataframe columns.
    It requires that that the user specifies the column name(s) to apply
    the check to. These must match the column names defined in the schema.

    Args:
        column_names (str | list[str]): The column name(s), defined in the
            schema, to apply the check to.

    Raises:
        TypeError: If `column_names` is not a `str` or `list[str]`.

    Returns:
        Callable: The decorated function.

    Examples:
        >>> from pandabear import DataFrameModel, Field, check
        >>>
        >>> class MySchema(DataFrameModel):
        >>>     column_a: int = Field()
        >>>     column_b: int = Field()
        >>>
        >>>     @check("column_a")
        >>>     def check_column(column: pd.Series) -> bool:
        >>>         return column.sum() > 0
        >>>
        >>>     # or
        >>>     @check(["column_a", "column_b"])
        >>>     def check_columns(column: pd.Series) -> bool:
        >>>         return column.sum() > 0
    """
    if not isinstance(column_names, (str, list)):
        raise TypeError(f"Expected `str` or `list[str]`, but found {type(column_names)}")

    column_names = [column_names] if type(column_names) == str else column_names

    def decorator(method: Callable) -> Callable:
        # Annotate method with check information. This allows the `validate`
        # method to find check functions and apply them to the correct columns.
        method.__setattr__("__check__", column_names)
        return method

    return decorator


def dataframe_check(method: Callable) -> Callable:
    """Decorator for defining custom checks on dataframes.

    This decorator is used to define custom checks on dataframes.

    Args:
        method (Callable): The method to decorate.

    Returns:
        Callable: The decorated function.

    Examples:
        >>> from pandabear import DataFrameModel, Field, dataframe_check
        >>>
        >>> class MySchema(DataFrameModel):
        >>>     column_a: int = Field()
        >>>     column_b: int = Field()
        >>>
        >>>     @dataframe_check
        >>>     def check_dataframe(df: pd.DataFrame) -> bool:
        >>>         return df.sum().sum() > 0
    """

    def wrapper(df: pd.DataFrame) -> bool:
        return method(df)

    # Annotate method with check information. This allows the `validate`
    # method to find check functions and apply them to the correct columns.
    wrapper.__check__ = None

    return wrapper



from typing import Any, Iterable

import pandas as pd


def series_greater_equal(series: pd.Series, value: Any) -> pd.Series:
    return series >= value


def series_greater(series: pd.Series, value: Any) -> pd.Series:
    return series > value


def series_less_equal(series: pd.Series, value: Any) -> pd.Series:
    return series <= value


def series_less(series: pd.Series, value: Any) -> pd.Series:
    return series < value


def series_isin(series: pd.Series, value: Iterable) -> pd.Series:
    return series.isin(value)


def series_notin(series: pd.Series, value: Iterable) -> pd.Series:
    return ~series.isin(value)


def series_str_contains(series: pd.Series, value: str) -> pd.Series:
    return series.str.contains(value)


def series_str_endswith(series: pd.Series, value: str) -> pd.Series:
    return series.str.endswith(value)


def series_str_startswith(series: pd.Series, value: str) -> pd.Series:
    return series.str.startswith(value)


def series_notnull(series: pd.Series, value: bool = True) -> pd.Series:
    return series.notnull()


def series_unique(series: pd.Series, value: bool = True) -> pd.Series:
    return ~series.duplicated(keep="first")


CHECK_NAME_FUNCTION_MAP = {
    "ge": series_greater_equal,
    "gt": series_greater,
    "le": series_less_equal,
    "lt": series_less,
    "isin": series_isin,
    "notin": series_notin,
    "str_contains": series_str_contains,
    "str_startswith": series_str_startswith,
    "str_endswith": series_str_endswith,
    "notnull": series_notnull,
    "unique": series_unique,
}



__version__ = "0.7.0"


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

from pandabear.decorators import check, check_schemas, dataframe_check
from pandabear.model import DataFrameModel, SeriesModel
from pandabear.model_components import Field, Index
from pandabear.typehints import DataFrame, Series
